# coding: utf8

"""T1 Linear - Clinica Visualizer.
This file has been generated automatically by the `clinica generate template`
command line tool. See here for more details:
https://aramislab.paris.inria.fr/clinica/docs/public/latest/InteractingWithClinica/
"""
